@@ -4,23 +4,28 @@
 
 """ Public XML parser module """
 
-import re
+import base64
+import hashlib
+import io
+import lxml.etree
 import os
-import codecs
+import re
+import requests
 import shutil
 import six
 import time
-import requests
-import lxml.etree
 import xml2rfc.log
 import xml2rfc.utils
 
+from xml2rfc.writers import base
+
 try:
-    from urllib.parse import urlparse, urljoin
+    from urllib.parse import urlparse, urljoin, urlsplit
 except ImportError:
-    from urlparse import urlparse, urljoin
+    from urlparse import urlparse, urljoin, urlsplit
+
 try:
-    import debug
+    from xml2rfc import debug
     assert debug
 except ImportError:
     pass
@@ -41,27 +46,34 @@ class XmlRfcError(Exception):
         self.filename = filename
         self.line = line_no
 
+    def __str__(self):
+        return self.msg
+
 class CachingResolver(lxml.etree.Resolver):
     """ Custom ENTITY request handler that uses a local cache """
     def __init__(self, cache_path=None, library_dirs=None, source=None,
-                 templates_path='templates', verbose=False, quiet=False,
-                 no_network=False, network_locs= [
+                 templates_path=base.default_options.template_dir, verbose=None, quiet=None,
+                 no_network=None, network_locs= [
+                     'https://bib.ietf.org/public/rfc/',
                      'https://xml2rfc.ietf.org/public/rfc/',
                      'https://xml2rfc.tools.ietf.org/public/rfc/',
+                     'http://bib.ietf.org/public/rfc/',
                      'http://xml2rfc.ietf.org/public/rfc/',
                      'http://xml2rfc.tools.ietf.org/public/rfc/',
                  ],
-                 rfc_number=None):
-        self.verbose = verbose
-        self.quiet = quiet
-        self.source = source
+                 rfc_number=None, options=base.default_options):
+        self.quiet = quiet if quiet != None else options.quiet
+        self.verbose = verbose if verbose != None else options.verbose
+        self.no_network = no_network if no_network != None else options.no_network
+        self.cache_path = cache_path or options.cache
+        self.source = source 
         self.library_dirs = library_dirs
         self.templates_path = templates_path
-        self.no_network=no_network
         self.network_locs = network_locs
         self.include = False
         self.rfc_number = rfc_number
         self.cache_refresh_secs = (60*60*24*14) # 14 days
+        self.options = options
 
         # Get directory of source
         if self.source:
@@ -75,9 +87,9 @@ class CachingResolver(lxml.etree.Resolver):
         # Determine cache directories to read/write to
         self.read_caches = [ os.path.expanduser(path) for path in xml2rfc.CACHES ]
         self.write_cache = None
-        if cache_path:
+        if self.cache_path:
             # Explicit directory given, set as first directory in read_caches
-            self.read_caches.insert(0, cache_path)
+            self.read_caches.insert(0, self.cache_path)
         # Try to find a valid directory to write to by stepping through
         # Read caches one by one
         for dir in self.read_caches:
@@ -112,8 +124,9 @@ class CachingResolver(lxml.etree.Resolver):
             path = os.path.join(dir, xml2rfc.CACHE_PREFIX)
             if os.access(path, os.W_OK):
                 shutil.rmtree(path)
-                xml2rfc.log.write('Deleted cache directory at', path)
+                xml2rfc.log.note('Deleted cache directory at', path)
 
+                
     def resolve(self, request, public_id, context):
         """ Called internally by lxml """
         if not request:
@@ -127,17 +140,28 @@ class CachingResolver(lxml.etree.Resolver):
             if self.rfc_number:
                 return self.resolve_string(self.rfc_number, context)
             return self.resolve_string("XXXX", context)
-        if not urlparse(request).netloc:
-            # Format the request from the relative path of the source so that 
-            # We get the exact same path as in the XML document
-            if request.startswith("file:///"):
-                request = request[8:]
-            try:
-                request = os.path.relpath(request, self.source_dir)
-            except ValueError:
-                pass
-        path = self.getReferenceRequest(request)
-        return self.resolve_filename(path, context)
+        # Warn if .ent file is referred.
+        if request.endswith('.ent'):
+            xml2rfc.log.warn('{} is no longer needed as the special processing of non-ASCII characters has been superseded by direct support for non-ASCII characters in RFCXML.'.format(request))
+        url = urlparse(request)
+        if not url.netloc or url.scheme == 'file':
+            if request.startswith("file://"):
+                request = request[7:]
+            if not self.options.allow_local_file_access:
+                path = self.getReferenceRequest(request)
+                abspath = os.path.dirname(os.path.abspath(path))
+                if not (abspath == self.options.template_dir or
+                        abspath == self.source_dir):
+                    error = 'Can not access local file: {}'.format(request)
+                    xml2rfc.log.error(error)
+                    raise XmlRfcError(error)
+        try:
+            path = self.getReferenceRequest(request)
+            return self.resolve_filename(path, context)
+        except Exception as e:
+            xml2rfc.log.error(str(e))
+            return None
+
     
     def getReferenceRequest(self, request, include=False, line_no=0):
         """ Returns the correct and most efficient path for an external request
@@ -181,7 +205,8 @@ class CachingResolver(lxml.etree.Resolver):
             request.  If not found, the request will be cached at write_dir.
 
             This method will throw an lxml.etree.XMLSyntaxError to be handled
-            by the application if the reference cannot be properly resolved
+            by the application if the reference or referencegroup cannot be
+            properly resolved
         """
         self.include = include # include state
         tried_cache = False
@@ -189,8 +214,12 @@ class CachingResolver(lxml.etree.Resolver):
         original = request  # Used for the error message only
         result = None  # Our proper path
         if request.endswith('.dtd') or request.endswith('.ent'):
-            if os.path.isabs(request) or urlparse(request).netloc:
-                # Absolute request, return as-is
+            if os.path.isabs(request) and os.path.exists(request):
+                # Absolute request, return as-is if it exists
+                attempts.append(request)
+                result = request
+            elif urlparse(request).netloc:
+                # Network request, return as-is
                 attempts.append(request)
                 result = request
             else:
@@ -205,10 +234,12 @@ class CachingResolver(lxml.etree.Resolver):
                     result = os.path.join(self.source_dir, basename)
                     attempts.append(result)
         else:
-            if request.endswith('.xml'):
+            if self.options and self.options.vocabulary == 'v3':
                 paths = [ request ]
-            else:
+            elif not request.endswith('.xml'):
                 paths = [ request, request + '.xml' ]
+            else:
+                paths = [ request ]                
             if os.path.isabs(paths[0]):
                 # Absolute path, return as-is
                 for path in paths:
@@ -236,8 +267,11 @@ class CachingResolver(lxml.etree.Resolver):
                         result = self.cache(path)
                         if result:
                             break
-                if not result and self.no_network:
-                    xml2rfc.log.warn("Document not found in cache, and --no-network specified -- couldn't resolve %s" % request)
+                if not result:
+                    if self.options.vocabulary == 'v3' and not request.endswith('.xml'):
+                        xml2rfc.log.warn("The v3 formatters require full explicit URLs of external resources.  Did you forget to add '.xml' (or some other extension)?")
+                    if self.no_network:
+                        xml2rfc.log.warn("Document not found in cache, and --no-network specified -- couldn't resolve %s" % request)
                 tried_cache = True
             else:
                 if os.path.dirname(paths[0]):
@@ -270,7 +304,6 @@ class CachingResolver(lxml.etree.Resolver):
                         #     result = os.path.join(self.source_dir, request)
                         #     attempts.append(result)
                 else:
-                    # Hanging filename
                     for dir in self.library_dirs:
                         # NOTE: Recursion can be implemented here
                         # Try local library directories
@@ -305,9 +338,9 @@ class CachingResolver(lxml.etree.Resolver):
         # Verify the result -- either raise exception or return it
         if not result or (not os.path.exists(result) and not urlparse(original).netloc):
             if os.path.isabs(original):
-                xml2rfc.log.warn('A reference was requested with an absolute path, but not found '
-                    'in that location.  Removing the path component will cause xml2rfc to look for'
-                    'the file automatically in standard locations.')
+                xml2rfc.log.warn('A reference was requested with an absolute path: "%s", but not found '
+                    'in that location.  Removing the path component will cause xml2rfc to look for '
+                    'the file automatically in standard locations.' % original)
             # Couldn't resolve.  Throw an exception
             error = XmlRfcError('Unable to resolve external request: '
                                       + '"' + original + '"', line_no=line_no, filename=self.source)
@@ -328,8 +361,10 @@ class CachingResolver(lxml.etree.Resolver):
 
             Checks for the existence of the cache and creates it if necessary.
         """
-        scheme, netloc, path, params, query, fragment = urlparse(url)
-        basename = os.path.basename(path)
+        scheme, netloc, path, query, fragment = urlsplit(url)
+        root, ext = os.path.splitext(path)
+        hash = '-'+base64.urlsafe_b64encode(hashlib.sha1(query.encode()).digest()).decode() if query else ''
+        basename = os.path.basename(root+hash+ext)
         typename = self.include and 'include' or 'entity'
         # Try to load the URL from each cache in `read_cache`
         for dir in self.read_caches:
@@ -341,21 +376,61 @@ class CachingResolver(lxml.etree.Resolver):
                 else:
                     xml2rfc.log.note('Resolving ' + typename + '...', url)
                     xml2rfc.log.note('Loaded from cache', cached_path)
-                    return cached_path
+                    xml = lxml.etree.parse(cached_path)
+                    if xml.getroot().tag == 'reference':
+                        if self.validate_ref(xml):
+                            return cached_path
+                        else:
+                            xml2rfc.log.error('Failure validating reference xml from %s' % cached_path )
+                            os.path.unlink(cached_path)
+                            return url
+                    elif xml.getroot().tag == 'referencegroup':
+                        if self.validate_ref_group(xml):
+                            return cached_path
+                        else:
+                            xml2rfc.log.error('Failure validating referencegroup xml from %s' % cached_path )
+                            os.path.unlink(cached_path)
+                    else:
+                        return cached_path
 
         xml2rfc.log.note('Resolving ' + typename + '...', url)
         if not netloc in self.sessions:
             self.sessions[netloc] = requests.Session()
         session = self.sessions[netloc]
-        r = session.get(url)
+        exc = None
+        for i in range(4):
+            try:
+                r = session.get(url)
+                break
+            except requests.exceptions.ConnectionError as e:
+                exc = e
+                xml2rfc.log.note('  retrying %s (%s)' % (url, e.args[0].args[0]))
+        else:
+            xml2rfc.log.error('Failure fetching URL %s (%s)' % (url, exc.args[0].args[0]))
+            return ''
+        for rr in r.history + [r, ]:
+            xml2rfc.log.note(' ... %s %s' % (rr.status_code, rr.url))
         if r.status_code == 200:
             if self.write_cache:
-                write_path = os.path.join(self.write_cache, 
-                                          xml2rfc.CACHE_PREFIX, basename)
-                with codecs.open(write_path, 'w', encoding='utf-8') as cache_file:
-                    cache_file.write(r.text)
-                xml2rfc.log.note('Added file to cache: ', write_path)
-                return write_path
+                try:
+                    xml = lxml.etree.fromstring(r.text.encode('utf8'))
+                    if xml.tag == 'reference':
+                        if self.validate_ref(xml):
+                            return self.add_to_cache(r.url, xml, basename)
+                        else:
+                            xml2rfc.log.error('Failure validating reference xml from %s' % url )
+                            return url
+                    elif xml.tag == 'referencegroup':
+                        if self.validate_ref_group(xml):
+                            return self.add_to_cache(r.url, xml, basename)
+                        else:
+                            xml2rfc.log.error('Failure validating referencegroup xml from %s' % url )
+                            return url
+                    else:
+                        return url
+                except Exception as e:
+                    xml2rfc.log.error(str(e))
+                    return url
             else:
                 return url
         else:
@@ -363,9 +438,30 @@ class CachingResolver(lxml.etree.Resolver):
             xml2rfc.log.note("URL retrieval failed with status code %s for '%s'" % (r.status_code, r.url))
             return ''
 
+    def validate_ref(self, xml):
+        ref_rng_file = os.path.join(os.path.dirname(__file__), 'data', 'reference.rng')
+        ref_rng = lxml.etree.RelaxNG(file=ref_rng_file)
+        return ref_rng.validate(xml)
+
+    def validate_ref_group(self, xml):
+        ref_rng_file = os.path.join(os.path.dirname(__file__), 'data', 'referencegroup.rng')
+        ref_rng = lxml.etree.RelaxNG(file=ref_rng_file)
+        return ref_rng.validate(xml)
+
+    def add_to_cache(self, url, xml, basename):
+        xml.set('{%s}base'%xml2rfc.utils.namespaces['xml'], url)
+        text = lxml.etree.tostring(xml, encoding='utf-8')
+        write_path = os.path.join(self.write_cache,
+                                  xml2rfc.CACHE_PREFIX, basename)
+        with io.open(write_path, 'w', encoding='utf-8') as cache_file:
+            cache_file.write(text.decode('utf-8'))
+        xml2rfc.log.note('Added file to cache: ', write_path)
+        return write_path
 
 class AnnotatedElement(lxml.etree.ElementBase):
     pis = None
+    page = None
+    outdent = 0
     def get(self, key, default=None):
         value = super(AnnotatedElement, self).get(key, default)
         if value == default:
@@ -376,39 +472,35 @@ class AnnotatedElement(lxml.etree.ElementBase):
 class XmlRfcParser:
 
     nsmap = {
-        'xi':   'http://www.w3.org/2001/XInclude',
+        b'xi':   b'http://www.w3.org/2001/XInclude',
     }
 
 
     """ XML parser container with callbacks to construct an RFC tree """
-    def __init__(self, source, verbose=False, quiet=False, options=None,
-                 cache_path=None, templates_path=None, library_dirs=None,
-                 no_network=False, network_locs=[
+    def __init__(self, source, verbose=None, quiet=None, options=base.default_options,
+                 cache_path=None, templates_path=base.default_options.template_dir, library_dirs=None, add_xmlns=False,
+                 no_network=None, network_locs=[
                      'https://xml2rfc.ietf.org/public/rfc/',
                      'https://xml2rfc.tools.ietf.org/public/rfc/',
                      'http://xml2rfc.ietf.org/public/rfc/',
                      'http://xml2rfc.tools.ietf.org/public/rfc/',
                  ]
                  ):
-        self.verbose = verbose
-        self.quiet = quiet
+        self.options = options
+        self.quiet = quiet if quiet != None else options.quiet
+        self.verbose = verbose if verbose != None else options.verbose
+        self.no_network = no_network if no_network != None else options.no_network
         self.source = source
-        self.cache_path = cache_path
-        self.no_network = no_network
+        self.cache_path = cache_path or options.cache
         self.network_locs = network_locs
 
         if self.source:
-            if six.PY2:
-                with open(self.source, "rU") as f:
-                    self.text = f.read()
-            else:
-                with open(self.source, "rb", newline=None) as f:
-                    self.text = f.read()
+            with io.open(self.source, "rb", newline=None) as f:
+                self.text = f.read()
 
         # Initialize templates directory
-        self.templates_path = templates_path or \
-                              os.path.join(os.path.dirname(xml2rfc.__file__),
-                                           'templates')
+        self.templates_path = templates_path
+
         if options and options.vocabulary == 'v2':
             self.default_dtd_path = os.path.join(self.templates_path, 'rfc2629.dtd')
         else:
@@ -422,12 +514,13 @@ class XmlRfcParser:
         #   2. Default to /usr/share/xml2rfc
         # Split on colon or semi-colon delimiters
         if not library_dirs:
-            library_dirs = os.environ.get('XML_LIBRARY', '/usr/share/xml2rfc:')
+            library_dirs = os.environ.get('XML_LIBRARY', '/usr/share/xml2rfc')
         self.library_dirs = []
-        for raw_dir in re.split(':|;', library_dirs):
+        srcdir = os.path.abspath(os.path.dirname(self.source)) if source else ''
+        for raw_dir in re.split(':|;', library_dirs) + [ srcdir ]:
             # Convert empty directory to source dir
             if raw_dir == '': 
-                raw_dir = os.path.abspath(os.path.dirname(self.source))
+                raw_dir = srcdir
             else:
                 raw_dir = os.path.normpath(os.path.expanduser(raw_dir))
             # Add dir if its unique
@@ -442,18 +535,28 @@ class XmlRfcParser:
                                         network_locs=self.network_locs,
                                         verbose=self.verbose,
                                         quiet=self.quiet,
+                                        options=options,
                                     )
 
     def delete_cache(self, path=None):
         self.cachingResolver.delete_cache(path=path)
 
-    def parse(self, remove_comments=True, remove_pis=False, quiet=False, strip_cdata=True, normalize=False):
+    def parse(self, remove_comments=True, remove_pis=False, quiet=False, strip_cdata=True, normalize=False, add_xmlns=False):
         """ Parses the source XML file and returns an XmlRfc instance """
-        if not (self.quiet or quiet):
-            xml2rfc.log.write('Parsing file', self.source)
+        xml2rfc.log.note('Parsing file', self.source)
+
+        # workaround for not being able to explicitly set namespaces on the
+        # xml root element in lxml: Insert it before we parse:
+        text = self.text
+        if add_xmlns and not self.nsmap[b'xi'] in text:
+            if re.search(b"""<rfc[^>]*(?P<xi>xmlns:xi=['"][^'"]+['"])""", text):
+                xml2rfc.log.error('Namespace prefix "xi" is defined, but not as the XInclude namespace "%s".' % self.nsmap[b'xi'])
+            else:
+                text = text.replace(b'<rfc ', b'<rfc xmlns:%s="%s" ' % (b'xi', self.nsmap[b'xi']), 1)
 
         # Get an iterating parser object
-        file = six.BytesIO(self.text)
+        file = six.BytesIO(text)
+        file.name = self.source
         context = lxml.etree.iterparse(file,
                                       dtd_validation=False,
                                       load_dtd=True,
@@ -476,23 +579,27 @@ class XmlRfcParser:
                                         network_locs=self.network_locs,
                                         verbose=self.verbose,
                                         quiet=self.quiet,
+                                        options=self.options,
                                      )
         context.resolvers.add(caching_resolver)
 
         # Get hold of the rfc number (if any) in the rfc element, so we can
         # later resolve the "&rfc.number;" entity.
         self.rfc_number = None
+        self.format_version = None
         try:
             for action, element in context:
                 if element.tag == "rfc":
                     self.rfc_number = element.attrib.get("number", None)
+                    self.format_version = element.attrib.get("version", None)
                     break
-        except lxml.etree.XMLSyntaxError as e:
-            xml2rfc.log.warn("Parsing Error: %s" % e)
         except ValueError as e:
             if e.message=="I/O operation on closed file":
                 pass
 
+        if self.format_version == "3":
+            self.default_dtd_path = None
+
         # now get a regular parser, and parse again, this time resolving entities
         parser = lxml.etree.XMLParser(dtd_validation=False,
                                       load_dtd=True,
@@ -500,7 +607,7 @@ class XmlRfcParser:
                                       no_network=self.no_network,
                                       remove_comments=remove_comments,
                                       remove_pis=remove_pis,
-                                      remove_blank_text=True,
+                                      remove_blank_text=False,
                                       resolve_entities=True,
                                       strip_cdata=strip_cdata)
 
@@ -514,6 +621,7 @@ class XmlRfcParser:
                                         verbose=self.verbose,
                                         quiet=self.quiet,
                                         rfc_number = self.rfc_number,
+                                        options=self.options,
                                     )
 
         # Add our custom resolver
@@ -525,7 +633,8 @@ class XmlRfcParser:
         parser.set_element_class_lookup(element_lookup)
 
         # Parse the XML file into a tree and create an rfc instance
-        file = six.BytesIO(self.text)
+        file = six.BytesIO(text)
+        file.name = self.source
         tree = lxml.etree.parse(file, parser)
         xmlrfc = XmlRfc(tree, self.default_dtd_path, nsmap=self.nsmap)
         xmlrfc.source = self.source
@@ -555,7 +664,7 @@ class XmlRfcParser:
                                                       no_network=False,
                                                       remove_comments=remove_comments,
                                                       remove_pis=remove_pis,
-                                                      remove_blank_text=True,
+                                                      remove_blank_text=False,
                                                       resolve_entities=True,
                                                       strip_cdata=strip_cdata)
                         parser.set_element_class_lookup(element_lookup)
@@ -587,7 +696,7 @@ class XmlRfcParser:
 
         return xmlrfc
 
-class XmlRfc:
+class XmlRfc(object):
     """ Internal representation of an RFC document
 
         Contains an lxml.etree.ElementTree, with some additional helper
@@ -597,11 +706,14 @@ class XmlRfc:
     """
 
     pis = {}
+    source = "(unknown source)"
 
-    def __init__(self, tree, default_dtd_path, nsmap={}):
+    def __init__(self, tree, default_dtd_path, source=None, nsmap={}):
         self.default_dtd_path = default_dtd_path
         self.tree = tree
         self.nsmap = nsmap
+        if source:
+            self.source = source
         # Pi default values
         self.pis = {
             "artworkdelimiter":	None,
