@@ -3,8 +3,8 @@
 # --------------------------------------------------
 
 # Python libs
-import datetime
 import calendar
+import six
 
 try:
     import debug
@@ -25,7 +25,7 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
     """
 
     def __init__(self, xmlrfc, width=72, quiet=None, options=default_options,
-                               date=datetime.date.today(), omit_headers=False):
+                               date=None, omit_headers=False):
         if not quiet is None:
             options.quiet = quiet
         RawTextRfcWriter.__init__(self, xmlrfc, options=options, date=date)
@@ -83,7 +83,7 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
             # Extend the number of lines if greater
             if need > needLines:
                 needLines = need
-        self.break_hints[where] = (int(needLines), type)
+        self.break_hints[where] = (needLines, type)
             
 
     def write_with_break_hint(self, writer, type, *args, **kwargs):
@@ -100,8 +100,14 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
 
     def needLines(self, count):
         """Deal with the PI directive needLines"""
+        if isinstance(count, six.string_types):
+            if count.isdigit():
+                count = int(count)
+            else:
+                xml2rfc.log.warn('Expected a numeric value for needLines, but got "%s".  Using 3.' % count)
+                count = 3
         try:
-            if int(count) < 0:
+            if count < 0:
                 self._set_break_hint(1, 'break', len(self.buf))
             else:
                 self._set_break_hint(count, 'raw', len(self.buf))
@@ -198,7 +204,7 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
             year = date.attrib.get('year', '')
             self.right_header = month + ' ' + year
         authors = self.r.findall('front/author')
-        surnames = [ a.attrib.get('surname') for a in authors if a.attrib.get('surname', None) ]
+        surnames = list(filter(None, [ a.get('asciiSurname', a.get('surname')) for a in authors ]))
         if len(surnames) == 1:
             self.left_footer = surnames[0]
         elif len(surnames) == 2:
