@@ -4,18 +4,18 @@
 
 # Python libs
 import textwrap
-import datetime
 import lxml
 import re
 
 try:
-    import debug
+    from xml2rfc import debug
     assert debug
 except ImportError:
     pass
 
 # Local lib
 from xml2rfc.writers.base import BaseRfcWriter, default_options
+from xml2rfc.util.unicode import expand_unicode_element
 import xml2rfc.utils
 
 
@@ -25,7 +25,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         The page width is controlled by the *width* parameter.
     """
 
-    def __init__(self, xmlrfc, width=72, margin=3, quiet=None, options=default_options, date=datetime.date.today()):
+    def __init__(self, xmlrfc, width=72, margin=3, quiet=None, options=default_options, date=None):
         if not quiet is None:
             options.quiet = quiet
         BaseRfcWriter.__init__(self, xmlrfc, options=options, date=date)
@@ -47,7 +47,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         self.iref_marks = {}
 
         # Text lookups
-        self.inline_tags = ['xref', 'eref', 'iref', 'cref', 'spanx']
+        self.inline_tags = ['xref', 'eref', 'iref', 'cref', 'spanx', 'u', ]
         
         # Custom textwrapper object
         self.wrapper = xml2rfc.utils.TextWrapper(width=self.width,
@@ -396,6 +396,14 @@ class RawTextRfcWriter(BaseRfcWriter):
         else:
             return target_text
 
+    def _expand_u(self, u):
+        try:
+            text = expand_unicode_element(u)
+        except (RuntimeError, ValueError) as exc:
+            text = ''
+            xml2rfc.log.error('%s'%exc)
+        return text
+
     def write_ref_element(self, key, text, sub_indent, source_line=None):
         """ Render a single reference element """
         # Use an empty first line if key is too long
@@ -447,26 +455,26 @@ class RawTextRfcWriter(BaseRfcWriter):
                     line.append('<' + element.attrib['target'].strip() + '>')
             elif element.tag == 'iref':
                 self._add_iref_to_index(element)
-            elif element.tag == 'cref' and \
-                element.pis['comments'] == 'yes':                
-                # Render if processing instruction is enabled
-                self.cref_counter += 1
-                anchor = element.attrib.get('anchor', None)
-                if anchor is None:
-                    anchor = 'CREF' + str(self.cref_counter)
-                    element.attrib['anchor'] = anchor
-                self._indexCref(self.cref_counter, anchor)
-                if element.pis['inline'] == 'yes':
-                    if anchor:
-                        anchor = anchor + ': '
-                    source = element.attrib.get('source', "")
-                    if source:
-                        source = " --" + source
-                    if element.text:
-                        line.append('[[' + anchor + element.text + source + ']]')
-                else:
-                    line.append('[' + anchor + ']')
-                    self.cref_list.append(element)
+            elif element.tag == 'cref':
+                if element.pis['comments'] == 'yes':                
+                    # Render if processing instruction is enabled
+                    self.cref_counter += 1
+                    anchor = element.attrib.get('anchor', None)
+                    if anchor is None:
+                        anchor = 'CREF' + str(self.cref_counter)
+                        element.attrib['anchor'] = anchor
+                    self._indexCref(self.cref_counter, anchor)
+                    if element.pis['inline'] == 'yes':
+                        if anchor:
+                            anchor = anchor + ': '
+                        source = element.attrib.get('source', "")
+                        if source:
+                            source = " --" + source
+                        if element.text:
+                            line.append('[[' + anchor + element.text + source + ']]')
+                    else:
+                        line.append('[' + anchor + ']')
+                        self.cref_list.append(element)
             elif element.tag == 'spanx':
                 style = element.attrib.get('style', 'emph')
                 edgechar = '_'  # default to emph because the spanx element exists
@@ -480,7 +488,11 @@ class RawTextRfcWriter(BaseRfcWriter):
                 if element.text:
                     text = element.text
                 line.append(edgechar + text + edgechar)
-            
+            elif element.tag == 'u':
+                line.append(self._expand_u(element))
+            else:
+                xml2rfc.log.error("Found unexpected inline element: <%s>" % element.tag)
+
             # Add tail text before next element
             if element.tail:
                 line.append(element.tail)
@@ -756,7 +768,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                     refstring.append(', '+seriesInfo.attrib['value'] + ' (work in progress)')
                 else:
                     refstring.append(', '+seriesInfo.attrib['name'] + u'\u00A0' +
-                                     seriesInfo.attrib['value'].replace('/', '/' + u'\u200B'))
+                                     seriesInfo.attrib['value'].replace('/', '/' + u'\uE060'))
             date = ref.find('front/date')
             if date is not None:
                 month = date.attrib.get('month', '')
@@ -845,7 +857,6 @@ class RawTextRfcWriter(BaseRfcWriter):
                 inline_text, null = \
                     self._combine_inline_elements(ttcol.getchildren())
                 text += inline_text
-            text = self.wrapper.replace(text)
             matrix[row].append(text)
 
         num_columns = len(matrix[0])
@@ -870,7 +881,6 @@ class RawTextRfcWriter(BaseRfcWriter):
                 inline_text, null = \
                     self._combine_inline_elements(cell.getchildren())
                 text += inline_text
-            text = self.wrapper.replace(text)
             matrix[row].append(text)
 
         # Get table style and determine maximum width of table
@@ -1425,6 +1435,9 @@ class RawTextRfcWriter(BaseRfcWriter):
                 line = line.replace(u'\u00A0', u' ')
                 line = line.replace(u'\u2011', u'-')
                 line = line.replace(u'\u200B', u'')
+                line = line.replace(u'\u2028', u' ')
+                line = line.replace(u'\u2060', u'')
+                assert line == line.replace(u'\uE060', u'')
             output.append(line);
         return output
 
