@@ -5,34 +5,28 @@
 import os
 import re
 import lxml.etree
-import datetime
 import traceback as tb
 
+from collections import OrderedDict
 from io import open
-from lxml import etree
 from lxml.etree import Element, Comment, CDATA
 
+import xml2rfc
 from xml2rfc import log
-from xml2rfc.utils import find_duplicate_ids, hastext, isempty
-from xml2rfc.writers.base import default_options
+from xml2rfc.util.unicode import unicode_content_tags, unicode_replacements, isascii
+from xml2rfc.utils import hastext, isempty, sdict, slugify, iscomment
+from xml2rfc.writers.base import default_options, BaseV3Writer
 
 
 try:
-    import debug
+    from xml2rfc import debug
     assert debug
 except ImportError:
     pass
 
-def slugify(s):
-    s = s.strip().lower()
-    s = re.sub(r'[^\w\s/|@=-]', '', s)
-    s = re.sub(r'[-_\s/|@=]+', '_', s)
-    s = s.strip('_')
-    return s
-
 def idref(s):
     s = re.sub(r'^[^A-Za-z_]', '_', s)
-    s = re.sub(r'[^-._A-Za-z0-9]', '-', s)
+    s = re.sub(r'[^-._A-Za-z0-9]', '_', s)
     return s
 
 def stripattr(e, attrs):
@@ -45,35 +39,20 @@ def copyattr(a, b):
         v = a.get(k)
         b.set(k, v)
 
-class V2v3XmlWriter(object):
+class V2v3XmlWriter(BaseV3Writer):
     """ Writes an XML file with v2 constructs converted to v3"""
 
-    def __init__(self, xmlrfc, quiet=None, options=default_options, date=datetime.date.today()):
+    def __init__(self, xmlrfc, quiet=None, options=default_options, date=None):
+        super(V2v3XmlWriter, self).__init__(xmlrfc, quiet=quiet, options=options, date=date)
         if not quiet is None:
             options.quiet = quiet
         self.xmlrfc = xmlrfc
         self.tree = xmlrfc.tree
         self.root = self.tree.getroot()
         self.options = options
-        self.v3_rng_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'v3.rng')
-        self.schema = etree.ElementTree(file=self.v3_rng_file)
-
-    def warn(self, e, text):
-        lnum = getattr(e, 'sourceline', 0)
-        msg = "%s(%s): Warning: %s" % (self.xmlrfc.source, lnum, text)
-        log.write(msg)
 
     def validate(self):
-        v3_rng_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'v3.rng')
-        v3_rng = lxml.etree.RelaxNG(file=v3_rng_file)
-        try:
-            v3_rng.assertValid(self.root)
-            log.note("The document validates according to the RFC7991 schema")
-        except Exception as e:
-            dups = find_duplicate_ids(self.schema, self.tree)
-            for attr, id, e in dups:
-                self.warn(e, 'Duplicate xsd:ID attribute %s="%s" found.  This will cause validation failure.' % (attr, id, ))
-            log.warn('\nInvalid document: %s' % (e,))
+        return super(V2v3XmlWriter, self).validate(when='when running v2v3 conversion', warn=True)
 
     def add_xinclude(self):
         for e in self.root.xpath('.//back//reference'):
@@ -85,7 +64,8 @@ class V2v3XmlWriter(object):
                         xi = self.element('{http://www.w3.org/2001/XInclude}include',
                                     nsmap=self.xmlrfc.nsmap,
                                     line=e.sourceline,
-                                    href="https://xml2rfc.tools.ietf.org/public/rfc/bibxml/reference.RFC.%s.xml"%num)
+                                    href="https://bib.ietf.org/public/rfc/bibxml/reference.RFC.%04d.xml"%int(num))
+                        xi.tail = e.tail
                         i = p.index(e)
                         p.remove(e)
                         p.insert(i, xi)
@@ -95,17 +75,26 @@ class V2v3XmlWriter(object):
                     name = si.get('value', '')
                     if name:
                         tag = name[len('draft-'):] if name.startswith('draft-') else name
-                        xi = self.element('{http://www.w3.org/2001/XInclude}include',
-                                    nsmap=self.xmlrfc.nsmap,
-                                    line=e.sourceline,
-                                    href="https://xml2rfc.tools.ietf.org/publicrfc/bibxml3/reference.I-D.%s.xml"%tag)
+                        if re.search(r'-\d\d$', tag) and self.options.draft_revisions:
+                            xi = self.element('{http://www.w3.org/2001/XInclude}include',
+                                        nsmap=self.xmlrfc.nsmap,
+                                        line=e.sourceline,
+                                        href="https://datatracker.ietf.org/doc/bibxml3/draft-%s.xml"%tag)
+                        else:
+                            tag = re.sub(r'-\d\d$', '', tag)
+                            xi = self.element('{http://www.w3.org/2001/XInclude}include',
+                                        nsmap=self.xmlrfc.nsmap,
+                                        line=e.sourceline,
+                                        href="https://datatracker.ietf.org/doc/bibxml3/draft-%s.xml"%tag)
+                        xi.tail = e.tail
                         i = p.index(e)
                         p.remove(e)
                         p.insert(i, xi)
                     else:
                         self.warn(e, 'Invalid value in %s' % lxml.etree.tostring(e))
-        lxml.etree.cleanup_namespaces(self.root, top_nsmap=self.xmlrfc.nsmap)
-
+#        lxml.etree.cleanup_namespaces(self.root, top_nsmap=self.xmlrfc.nsmap)
+                else:
+                    self.note(e, "Not implemented: xi:include handling for %s:%s" % (si.get('name'), si.get('value')))
 
     def post_process_lines(self, lines):
         output = [ line.replace(u'\u00A0', ' ') for line in lines ]
@@ -115,30 +104,39 @@ class V2v3XmlWriter(object):
         """ Public method to write the XML document to a file """
 
         self.convert2to3()
-        self.validate()
         if self.options.add_xinclude:
             self.add_xinclude()
 
+        # Use an internal DTD to declare entities
+        doctype_string = """<!DOCTYPE rfc [
+  <!ENTITY nbsp    "&#160;">
+  <!ENTITY zwsp   "&#8203;">
+  <!ENTITY nbhy   "&#8209;">
+  <!ENTITY wj     "&#8288;">
+]>"""
         # Use lxml's built-in serialization
         file = open(filename, 'w', encoding='utf-8')
         text = lxml.etree.tostring(self.root.getroottree(), 
-                                       xml_declaration=True, 
-                                       encoding='utf-8',
-                                       pretty_print=True)
-        file.write(text.decode('utf-8'))
+                                   encoding='unicode',
+                                   doctype=doctype_string,
+                                   pretty_print=True)
+
+        file.write(u"<?xml version='1.0' encoding='utf-8'?>\n")
+        file.write(text)
 
         if not self.options.quiet:
-            log.write('Created file', filename)
+            self.log(' Created file %s' % filename)
 
     # --- Element Operations -------------------------------------------
 
     def element(self, tag, line=None, **kwargs):
-        e = Element(tag, **kwargs)
-        if self.options.debug:
-            filename, lineno, caller, code = tb.extract_stack()[-2]
-            e.set('line', str(lineno))
+        e = Element(tag, **sdict(kwargs))
         if line:
             e.sourceline = line
+        elif self.options.debug:
+            filename, lineno, caller, code = tb.extract_stack()[-2]
+            e.base = os.path.basename(filename)
+            e.sourceline = lineno
         return e
 
     def copy(self, e, tag):
@@ -167,6 +165,7 @@ class V2v3XmlWriter(object):
             if self.options.verbose:
                 for comment in comments:
                     c = Comment(" v2v3: %s " % comment.strip())
+                    c.tail = ''
                     p.insert(i, c)
                     i += 1
             if not b is None:
@@ -174,19 +173,23 @@ class V2v3XmlWriter(object):
                     b.text = a.text
                 if a.tail != None:
                     b.tail = a.tail
-                b.sourceline = a.sourceline
+                if a.sourceline:
+                    b.sourceline = a.sourceline
                 copyattr(a, b)
                 for child in a.iterchildren():
                     b.append(child)         # moves child from a to b
                 p.replace(a, b)
             else:
-                if a.tail:
-                    if c is None:
-                        p.text = p.text + ' ' + a.tail if p.text else a.tail
-                    else:
-                        c.tail = a.tail
+                if iscomment(a):
+                    a.text = ''
+                for text in [a.text, a.tail]:
+                    if text:
+                        if c is None:
+                            p.text = p.text + text if p.text else text
+                        else:
+                            c.tail += text
                 p.remove(a)
-        if b != None:
+        if b != None and a.sourceline:
             b.sourceline = a.sourceline
         return b
 
@@ -239,7 +242,8 @@ class V2v3XmlWriter(object):
         if t is None:
             t = self.element('t', line=e.sourceline)
         t.text = e.text
-        t.sourceline = e.sourceline
+        if e.sourceline:
+            t.sourceline = e.sourceline
         e.text = None
         for c in e.iterchildren():
             t.append(c)
@@ -249,6 +253,10 @@ class V2v3XmlWriter(object):
     # ------------------------------------------------------------------
 
     def convert2to3(self):
+        if self.root.get('version') in ['3', ]:
+            return self.tree
+        log.note(' Converting v2 to v3: %s' % self.xmlrfc.source)
+
         selectors = [
             # we need to process list before block elements that might get
             # promoted out of their surrounding <t/>, as <list/> uses one <t/>
@@ -261,6 +269,8 @@ class V2v3XmlWriter(object):
                                             # 2.5.8.  "width" Attribute
                                             # 2.5.9.  "xml:space" Attribute
             './/back',
+            './/code',
+            './/date',
             # We need to process preamble and postamble before figure,
             # because artwork or sourcecode within a figure could later be
             # promoted and the figure discarded.
@@ -273,6 +283,7 @@ class V2v3XmlWriter(object):
                                             # 2.25.5.  "src" Attribute
                                             # 2.25.6.  "suppress-title" Attribute
                                             # 2.25.8.  "width" Attribute
+            './/relref',                    # Deprecated after 7991
             './/reference',                 #        <reference>
             '.',                            # 2.45.  <rfc>
                                             # 2.45.1.  "category" Attribute
@@ -280,7 +291,8 @@ class V2v3XmlWriter(object):
                                             # 2.45.3.  "docName" Attribute
                                             # 2.45.7.  "number" Attribute
                                             # 2.45.10.  "seriesNo" Attribute
-            './/seriesInfo',                # 2.47.  <seriesInfo>
+            #Disabled 
+            #'.//seriesInfo',                # 2.47.  <seriesInfo>
             './/t',                         # 2.53.  <t>
                                             # 2.53.2.  "hangText" Attribute
             './/xref',                      # 2.66.  <xref>
@@ -299,16 +311,16 @@ class V2v3XmlWriter(object):
                                             # 2.46.4.  "title" Attribute
             './/*[@anchor]',
             './/xref[@target]',
-            './/processing-instruction()',  # 1.3.2
+            '//processing-instruction()',  # 1.3.2
             # handle mixed block/non-block content surrounding all block nodes
             './/*[self::artwork or self::dl or self::figure or self::ol or self::sourcecode or self::t or self::ul]',
-            './/*[@*="yes" or @*="no"]',      # convert old attribute false/true
+            '//*[@*="yes" or @*="no"]',      # convert old attribute false/true
             '.;pretty_print_prep()',
+            '.;wrap_non_ascii()',
         ]
 
-        # replace the vocabulary v2 dtd, but keep some entity definitions.
-        tree = self.tree
-        tree.docinfo.system_url = "rfc2629-xhtml.ent"
+        # Remove any DOCTYPE declaration
+        self.tree.docinfo.clear()
 
         for s in selectors:
             slug = slugify(s.replace('self::', '').replace(' or ','_').replace(';','_'))
@@ -331,6 +343,21 @@ class V2v3XmlWriter(object):
             else:
                 log.warn("No handler for function %s, slug %s" % (func_name, slug, ))
 
+        self.root.set('version', '3')
+
+        # Add a comment about the converter version
+        conversion_version = Comment(' xml2rfc v2v3 conversion %s ' % xml2rfc.__version__)
+        conversion_version.tail = '\n  '
+        self.root.insert(0, conversion_version)
+
+        # This is a workaround for not being able to do anything about
+        # namespaces other than when creating an element.  It lets us retain
+        # a namespace declaration for xi: in the root element.
+#         dummy = self.element('{http://www.w3.org/2001/XInclude}include', nsmap=self.xmlrfc.nsmap)
+#         self.root.insert(0, dummy)
+#         lxml.etree.cleanup_namespaces(self.root, top_nsmap=self.xmlrfc.nsmap, keep_ns_prefixes='xi')
+#         self.root.remove(dummy)
+        log.note(' Completed v2 to v3 conversion')
         return self.tree
 
     # ----------------------------------------------------------------------
@@ -341,11 +368,13 @@ class V2v3XmlWriter(object):
     #       to <rfc> to cover Processing Instructions (PIs) that were in v2
     #       that are still needed in the grammar.  ...
     def processing_instruction(self, e, p):
+        if e.target != 'rfc':
+            return
         rfc_element = self.root.find('.')
         pi_name = {
             'sortrefs': 'sortRefs',
             'symrefs':  'symRefs',
-            'docdepth': 'tocDepth',
+            'tocdepth': 'tocDepth',
             'toc':      'tocInclude',
         }
         for k, a in pi_name.items():
@@ -353,8 +382,11 @@ class V2v3XmlWriter(object):
                 v = e.get(k)
                 if v == 'yes':
                     v = 'true'
+                elif v == 'no':
+                    v = 'false'
                 rfc_element.set(a, v)
                 self.replace(e, None, 'Moved %s PI to <rfc %s="%s"' % (k, a, v))
+                break
         else:
             self.replace(e, None, 'Removed %s PI"' % (k,))
 
@@ -404,7 +436,7 @@ class V2v3XmlWriter(object):
     # 
     #    Deprecated.
     def element_artwork(self, e, p):
-        if e.text:
+        if e.text and e.text.strip() and not ']]>' in e.text:
             e.text = CDATA(e.text)          # prevent text from being mucked up by other processors
         stripattr(e, ['height', '{http://www.w3.org/XML/1998/namespace}space', 'width', ])
         if e.text and re.search(r'^\s*<CODE BEGINS>', e.text):
@@ -426,9 +458,23 @@ class V2v3XmlWriter(object):
             if ends and tail.strip() != "":
                 self.warn(e, "Found non-whitespace content after <CODE ENDS>")
                 e.tail = tail
-            stripattr(e, ['align', 'height', 'suppress-title', 'width', ])
-            if self.options.strict:
-                stripattr(e, ['alt', ])
+            stripattr(e, ['align', 'alt', 'height', 'suppress-title', 'width', ])
+        src = e.get('src')
+        if e.text and e.text.strip() and src:
+            # We have both text content and a src attribute -- convert to
+            # <artset> with 2 <artwork)
+            ext = os.path.splitext(src)[1][1:]
+            artset = self.element('artset', line=e.sourceline)
+            extart = self.copy(e, 'artwork')
+            extart.text = None
+            extart.tail = None
+            extart.set('type', ext)
+            artset.append(extart)
+            stripattr(e, ['src'])
+            e.set('type', 'ascii-art')
+            artset.append(e)
+            p.append(artset)
+
 
     def element_back(self, e, p):
         # XXXX The RFCs don't say anything about the text rendering of
@@ -446,6 +492,18 @@ class V2v3XmlWriter(object):
                 refs.append(r)          # moves r
 
 
+    def element_code(self, e, p):
+        if e.text and re.search(r'^[A-Z][A-Z]?-', e.text):
+            cc, num = e.text.split('-', 1)
+            if cc in ['AX', 'CH', 'FI', 'HR', 'FL', 'LT', 'L', 'MC', 'MD', 'SE', 'SI', ]:
+                e.text = num
+
+    def element_date(self, e, p):
+        year = e.get('year')
+        if year and not year.isdigit():
+            del e.attrib['year']
+            e.text = year
+
     # 2.25.  <figure>
     # 
     # 2.25.1.  "align" Attribute
@@ -485,7 +543,7 @@ class V2v3XmlWriter(object):
     # 
     def element_figure(self, e, p):
         comments = []
-        embedded = e.find('./artwork')
+        embedded = e.find('.//artwork')
         if embedded == None:
             embedded = e.find('./sourcecode')
         if embedded != None:
@@ -506,8 +564,11 @@ class V2v3XmlWriter(object):
             and e.find('name')==None and e.find('preamble')==None and e.find('postamble')==None ):
             pos = p.index(e)
             embedded.tail = e.tail
+            children = e.getchildren()
             p.remove(e)
-            p.insert(pos, embedded)
+            for c in children:
+                p.insert(pos, c)
+                pos += 1
         else:
             stripattr(e, ['align', 'height', 'src', 'suppress-title', 'width', ])
             if self.options.strict:
@@ -551,6 +612,12 @@ class V2v3XmlWriter(object):
                 e.set('quoteTitle', v)
         stripattr(e, ['quote-title'])
 
+    def element_relref(self, e, p):
+        if 'displayFormat' in e.attrib:
+            e.attrib['sectionFormat'] = e.attrib['displayFormat']
+            del e.attrib['displayFormat']
+        e.tag = 'xref'
+
     # 2.45.  <rfc>
     # 
     # 2.45.1.  "category" Attribute
@@ -583,11 +650,13 @@ class V2v3XmlWriter(object):
 #             "info": "Informational",
 #         }
         def equal(e1, e2):
-            return etree.tostring(e1) == etree.tostring(e2)
+            return (e1.get('name'), e1.get('value')) ==  (e2.get('name'), e2.get('value'))
+        if e.get('ipr') == 'none':
+            return
         front = e.find('./front')
         title = e.find('./front/title')
         i = front.index(title) + 1 if title!=None else 0
-        series = front.xpath('seriesInfo')
+        series = front.xpath('seriesInfo') if front!=None else []
         if 'category' in e.attrib and 'seriesNo' in e.attrib:
             attr = {
                 'name': e.get('category'),
@@ -602,13 +671,9 @@ class V2v3XmlWriter(object):
             if not [ s for s in series if equal(s, new) ]:
                 front.insert(i, new)
             if 'docName' in e.attrib:
-                e.insert(0, self.element('link', rel='convertedFrom', href="https://datatracker.ietf.org/doc/%s"%(e.get('docName'), ), line=e.sourceline))
+                e.insert(0, self.element('link', rel='prev', href="https://datatracker.ietf.org/doc/%s"%(e.get('docName'), ), line=e.sourceline))
         elif 'docName' in e.attrib:
             value=e.get('docName')
-            if '.' in value:
-                log.warn("The 'docName' attribute of the <rfc/> element should not contain any filename extension: docName=\"draft-foo-bar-02\".")
-            if not re.search('-\d\d$', value):
-                log.warn("The 'docName' attribute of the <rfc/> element should have a revision number as the last component when submitted: docName=\"draft-foo-bar-02\".")
             new = self.element('seriesInfo', name="Internet-Draft", value=value, line=e.sourceline)
             if not 'Internet-Draft' in [ s.get('name') for s in series ]:
                 front.insert(i, new)
@@ -649,8 +714,6 @@ class V2v3XmlWriter(object):
     #       <xref> because it makes no sense semantically.
     #
     def element_xref(self, e, p):
-        if 'format' in e.attrib and e.get('format') == 'none':
-            stripattr(e, ['format'])
         stripattr(e, ['pageno'])
 
     # 2.66.2.  "pageno" Attribute
@@ -679,6 +742,7 @@ class V2v3XmlWriter(object):
     # 
     #    This element appears as a child element of <address> (Section 2.2).
     def element_facsimile(self, e, p):
+            e.text = ''
             self.replace(e, None)
 
     # 3.3.  <format>
@@ -694,8 +758,11 @@ class V2v3XmlWriter(object):
         if ptarget:
             if ftarget == ptarget:
                 self.replace(e, None, "<format/> element with duplicate target (%s) removed" % ftarget)
-            else:
-                self.replace(e, None, "Warning: <format/> element with alternative target (%s) removed" % ftarget)
+            # The following, while implementing RFC 7991, seems dubious, and
+            # has been commented out.  Why would you disallow pointing to
+            # multiple versions of a reference?
+            #else:
+            #    self.replace(e, None, "Warning: <format/> element with alternative target (%s) removed" % ftarget)
         else:
             p.set('target', ftarget)
             self.replace(e, None, "<format/> element removed, target value (%s) moved to parent" % ftarget)
@@ -724,7 +791,7 @@ class V2v3XmlWriter(object):
         # convert to dl, ul, or ol
         nstyle = None
         style = e.get('style', '').strip()
-        attribs = {}
+        attribs = OrderedDict()
         comments = []
         if not style:
             # otherwise look for the nearest list parent with a style and use it
@@ -770,7 +837,7 @@ class V2v3XmlWriter(object):
             comments.append("Warning: unknown list style: '%s'" % style)
         #
         comments.append('Replaced <list style="%s"/> with <%s/>' % (style, tag))
-        if 'counter' in e.keys():
+        if tag=='ol' and 'counter' in e.keys():
             attribs['group'] = e.get('counter')
             comments.append("Converting <list counter=...> to <%s group=...> " % tag)
         #
@@ -794,16 +861,16 @@ class V2v3XmlWriter(object):
                     del t.attrib['hangText']
                 # Convert <vspace> at the start of hanging list text to
                 # attribute hanging='true' on <dl>
-                if not t.text:
-                    if t[0].tag == 'vspace':
+                if not t.text or not t.text.strip():
+                    if len(t) and t[0].tag == 'vspace':
                         t.text = t[0].tail
                         t.remove(t[0])
-                    l.set('newline', 'true')
+                        l.set('newline', 'true')
                 i = l.index(t)
                 l.insert(i, dt)
                 self.replace(t, 'dd')
         else:
-            self.error("Unexpected tag when processing <list/>: '%s'" % tag)
+            self.warn(e, "Unexpected tag when processing <list/>: '%s'" % tag)
 
     # 3.4.1.  "counter" Attribute
     # 
@@ -828,6 +895,7 @@ class V2v3XmlWriter(object):
     #    and <texttable> (Section 3.8).
     def element_postamble(self, e, p):
         e = self.replace(e, 't')
+        e.set('keepWithPrevious', 'true')
         self.move_after(e, p)
 
     # 3.6.  <preamble>
@@ -839,6 +907,7 @@ class V2v3XmlWriter(object):
     #    and <texttable> (Section 3.8).
     def element_preamble(self, e, p):
         e = self.replace(e, 't')
+        e.set('keepWithNext', 'true')
         s = p.getprevious()
         if not s is None:
             self.move_after(e, s)
@@ -885,7 +954,8 @@ class V2v3XmlWriter(object):
         if style in tags:
             tag = tags[style]
         else:
-            self.error("Unexpected style in <spanx/>: '%s'" % style)
+            self.warn(e, "Unexpected style in <spanx/>: '%s'" % style)
+            tag = 'em'
         stripattr(e, ['style', '{http://www.w3.org/XML/1998/namespace}space', ])
         self.replace(e, tag, 'Replaced <spanx style="%s"/> with <%s/>' % (style, tag))
 
@@ -1026,6 +1096,7 @@ class V2v3XmlWriter(object):
                 l.remove(t)
         else:
             self.replace(e, None, "<vspace/> deprecated and removed")
+            self.warn(e, "Deprecated <vspace/> element removed, but no good conversion found  The output will most likely need fixup.")
 
     # 3.10.1.  "blankLines" Attribute
     # 
@@ -1064,17 +1135,67 @@ class V2v3XmlWriter(object):
             elif v == 'no':
                 e.set(k, 'false')
 
-    def pretty_print_prep(self, e, p):
-        # apply this to elements that can't appear with text, i.e., don't have
-        # any of these as parent:
-        skip_parents = set([
-            "annotation", "blockquote", "preamble", "postamble", "name", "refcontent", "c", "t",
-            "cref", "dd", "dt", "li", "td", "th", "tt", "em", "strong", "sub", "sup", ])
-        for c in e.iter():
-            p = c.getparent()
-            if p != None and p.tag in skip_parents:
-                continue
-            if c.tail != None:
-                if c.tail.strip() == '':
-                    c.tail = None
+#     def pretty_print_prep(self, e, p):
+#         # apply this to elements that can't appear with text, i.e., don't have
+#         # any of these as parent:
+#         skip_parents = set([
+#             "annotation", "blockquote", "preamble", "postamble", "name", "refcontent", "c", "t",
+#             "cref", "dd", "dt", "li", "td", "th", "tt", "em", "strong", "sub", "sup", ])
+#         for c in e.iter():
+#             p = c.getparent()
+#             if p != None and p.tag in skip_parents:
+#                 continue
+#             if c.tail != None:
+#                 if c.tail.strip() == '':
+#                     c.tail = None
 
+    def wrap_non_ascii(self, e, p):
+        self.downcode(replacements=unicode_replacements)
+        self.downcode_punctuation()
+        for e in self.tree.iter():
+            def uwrap(text, line): 
+                words = []
+                elements = []
+                ascii = None
+                for word in re.split('(\s+)', text, flags=re.U):
+                    if isascii(word):
+                        words.append(word)
+                    else:
+                        u = self.element('u', line=line)
+                        u.text = word
+                        if ascii is None:
+                            ascii = ''.join(words)
+                            words = []
+                            elements.append(u)
+                        else:
+                            elements[-1].tail = ''.join(words)
+                            words = []
+                            elements.append(u)
+                if words:
+                    if ascii is None:
+                        ascii = ''.join(words)
+                    else:
+                        elements[-1].tail = ''.join(words)
+                return ascii, elements
+            #
+            if e.text and e.tag not in unicode_content_tags:
+                    try:
+                        e.text.encode('ascii')
+                    except UnicodeEncodeError:
+                        e.text, elements = uwrap(e.text, e.sourceline)
+                        if len(e):
+                            c = e[0]
+                            for u in elements:
+                                c.addprevious(u)
+                        else:
+                            for u in elements:
+                                e.append(u)
+            if e.tail and e.getparent().tag not in unicode_content_tags:
+                    try:
+                        e.tail.encode('ascii')
+                    except UnicodeEncodeError:
+                        e.tail, elements = uwrap(e.tail, e.sourceline)
+                        for u in elements:
+                            e.addnext(u)
+                            e = u
+                                
